# date4/format/strptime.rb: Written by Tadayoshi Funaba 1999-2006
# $Id: strptime.rb,v 1.5 2006-09-24 07:58:38+09 tadf Exp $

class Date4

  def self.num_pattern? (s)
    /\A%[EO]?[CDdeFGgHIjkLlMmNQRrSsTUuVvWwXxYy\d]/ =~ s || /\A\d/ =~ s
  end

  private_class_method :num_pattern?

  def self._strptime_i(str, fmt, e)
    fmt.scan(/%[EO]?(:{1,3}z|.)|(.)/m) do |s, c|
      if s
	case s
	when 'A', 'a'
	  return unless str.sub!(/\A(#{e._lc.days.keys.join('|')})/i, '') ||
			str.sub!(/\A(#{e._lc.abbr_days.keys.join('|')})/i, '') ||
			str.sub!(/\A(#{e._lc_c.days.keys.join('|')})/i, '') ||
			str.sub!(/\A(#{e._lc_c.abbr_days.keys.join('|')})/i, '')
	  val = e._lc.days[$1.downcase] || e._lc.abbr_days[$1.downcase] ||
		e._lc_c.days[$1.downcase] || e._lc_c.abbr_days[$1.downcase]
	  return unless val
	  e.wday = val
	when 'B', 'b', 'h'
	  return unless str.sub!(/\A(#{e._lc.months.keys.join('|')})/i, '') ||
			str.sub!(/\A(#{e._lc.abbr_months.keys.join('|')})/i, '') ||
			str.sub!(/\A(#{e._lc_c.months.keys.join('|')})/i, '') ||
			str.sub!(/\A(#{e._lc_c.abbr_months.keys.join('|')})/i, '')
	  val = e._lc.months[$1.downcase] || e._lc.abbr_months[$1.downcase] ||
		e._lc_c.months[$1.downcase] || e._lc_c.abbr_months[$1.downcase]
	  return unless val
	  e.mon = val
	when 'C'
	  return unless str.sub!(if num_pattern?($')
				 then /\A([-+]?\d{1,2})/
				 else /\A([-+]?\d{1,})/
				 end, '')
	  val = $1.to_i
	  e._cent = val
	when 'c'
	  return unless _strptime_i(str, e._lc.c_fmt, e)
	when 'D'
	  return unless _strptime_i(str, '%m/%d/%y', e)
	when 'd', 'e'
	  return unless str.sub!(/\A( \d|\d{1,2})/, '')
	  val = $1.to_i
	  return unless (1..31) === val
	  e.mday = val
	when 'F'
	  return unless _strptime_i(str, '%Y-%m-%d', e)
	when 'G'
	  return unless str.sub!(if num_pattern?($')
				 then /\A([-+]?\d{1,4})/
				 else /\A([-+]?\d{1,})/
				 end, '')
	  val = $1.to_i
	  e.cwyear = val
	when 'g'
	  return unless str.sub!(/\A(\d{1,2})/, '')
	  val = $1.to_i
	  return unless (0..99) === val
	  e.cwyear = val
	  e._cent ||= if val >= 69 then 19 else 20 end
	when 'H', 'k'
	  return unless str.sub!(/\A( \d|\d{1,2})/, '')
	  val = $1.to_i
	  return unless (0..24) === val
	  e.hour = val
	when 'I', 'l'
	  return unless str.sub!(/\A( \d|\d{1,2})/, '')
	  val = $1.to_i
	  return unless (1..12) === val
	  e.hour = val
	when 'j'
	  return unless str.sub!(/\A(\d{1,3})/, '')
	  val = $1.to_i
	  return unless (1..366) === val
	  e.yday = val
	when 'L'
	  return unless str.sub!(if num_pattern?($')
				 then /\A([-+]?\d{1,3})/
				 else /\A([-+]?\d{1,})/
				 end, '')
#	  val = $1.to_i.to_r / (10**3)
	  val = $1.to_i.to_r / (10**$1.size)
	  e.sec_fraction = val
	when 'M'
	  return unless str.sub!(/\A(\d{1,2})/, '')
	  val = $1.to_i
	  return unless (0..59) === val
	  e.min = val
	when 'm'
	  return unless str.sub!(/\A(\d{1,2})/, '')
	  val = $1.to_i
	  return unless (1..12) === val
	  e.mon = val
	when 'N'
	  return unless str.sub!(if num_pattern?($')
				 then /\A([-+]?\d{1,9})/
				 else /\A([-+]?\d{1,})/
				 end, '')
#	  val = $1.to_i.to_r / (10**9)
	  val = $1.to_i.to_r / (10**$1.size)
	  e.sec_fraction = val
	when 'n', 't'
	  return unless _strptime_i(str, "\s", e)
	when 'P', 'p'
	  return unless str.sub!(/\A(#{e._lc.ampm.keys.join('|')})/i, '') ||
			str.sub!(/\A(#{e._lc_c.ampm.keys.join('|')})/i, '') ||
			str.sub!(/\A([ap])(?:m\b|\.m\.)/i, '')
	  val = e._lc.ampm[$1.downcase] || e._lc_c.ampm[$1.downcase] ||
	    if $1.downcase == 'a' then 0 else 12 end
	  return unless val
	  e._merid = val
	when 'Q'
	  return unless str.sub!(/\A(-?\d{1,})/, '')
	  val = $1.to_i.to_r / (10**3)
	  e.seconds = val
	when 'R'
	  return unless _strptime_i(str, '%H:%M', e)
	when 'r'
	  return unless _strptime_i(str, e._lc.r_fmt, e)
	when 'S'
	  return unless str.sub!(/\A(\d{1,2})/, '')
	  val = $1.to_i
	  return unless (0..60) === val
	  e.sec = val
	when 's'
	  return unless str.sub!(/\A(-?\d{1,})/, '')
	  val = $1.to_i
	  e.seconds = val
	when 'T'
	  return unless _strptime_i(str, '%H:%M:%S', e)
	when 'U', 'W'
	  return unless str.sub!(/\A(\d{1,2})/, '')
	  val = $1.to_i
	  return unless (0..53) === val
	  e.__send__(if s == 'U' then :wnum0= else :wnum1= end, val)
	when 'u'
	  return unless str.sub!(/\A(\d{1})/, '')
	  val = $1.to_i
	  return unless (1..7) === val
	  e.cwday = val
	when 'V'
	  return unless str.sub!(/\A(\d{1,2})/, '')
	  val = $1.to_i
	  return unless (1..53) === val
	  e.cweek = val
	when 'v'
	  return unless _strptime_i(str, '%e-%b-%Y', e)
	when 'w'
	  return unless str.sub!(/\A(\d{1})/, '')
	  val = $1.to_i
	  return unless (0..6) === val
	  e.wday = val
	when 'X'
	  return unless _strptime_i(str, e._lc.X_fmt, e)
	when 'x'
	  return unless _strptime_i(str, e._lc.x_fmt, e)
	when 'Y'
	  return unless str.sub!(if num_pattern?($')
				 then /\A([-+]?\d{1,4})/
				 else /\A([-+]?\d{1,})/
				 end, '')
	  val = $1.to_i
	  e.year = val
	when 'y'
	  return unless str.sub!(/\A(\d{1,2})/, '')
	  val = $1.to_i
	  return unless (0..99) === val
	  e.year = val
	  e._cent ||= if val >= 69 then 19 else 20 end
	when 'Z', /\A:{0,3}z/
	  return unless str.sub!(/\A((?:gmt|utc)?[-+]\d+(?:[,.:]\d+(?::\d+)?)?
				    |[a-z.\s]+(?:standard|daylight)\s+time\b
				    |[a-z]+(?:\s+dst)?\b
				    )/ix, '')
	  val = $1
	  e.zone = val
	  offset = zone_to_diff(val)
	  e.offset = offset
	when '%'
	  return unless str.sub!(/\A%/, '')
	when '+'
	  return unless _strptime_i(str, '%a %b %e %H:%M:%S %Z %Y', e)
	else
	  return unless str.sub!(Regexp.new('\\A' + Regexp.quote($1)), '')
	end
      else
	case c
	when /\A[\s\v]/
	  str.sub!(/\A[\s\v]+/, '')
	else
	  return unless str.sub!(Regexp.new('\\A' + Regexp.quote(c)), '')
	end
      end
    end
  end

  private_class_method :_strptime_i

  def self._strptime(*args)
    str, fmt, lc, = parsearg([[nil, :d, :date],
			      ['%F', :fmt, :format],
			      [Format::Locale::C, :lc, :locale]], args)
    e = Format::Bag.new
    e._lc = lc
    e._lc_c = Format::Locale::C

    return unless _strptime_i(str.dup, fmt, e)

    if e._cent
      if e.cwyear
	e.cwyear += e._cent * 100
      end
      if e.year
	e.  year += e._cent * 100
      end
    end

    if e._merid
      if e.hour
	e.hour %= 12
	e.hour += e._merid
      end
    end

    e.to_hash
  end

end

class DateTime4 < Date4

  def self._strptime(*args)
    str, fmt, lc, = parsearg([[nil, :d, :date],
			      ['%FT%T%z', :fmt, :format],
			      [Format::Locale::C, :lc, :locale]], args)
    super(str, fmt, lc)
  end

end
